% Copyright (c) 2011 Joris Degroote, Majid Hojjat, Electra Stavropoulou,
% Roland Wüchner, Kai-Uwe Bletzinger
%
% Licensed under the Non-Profit Open Software License version 3.0
% (NPOSL-3.0). You should have received a copy of the NPOSL-3.0 along with 
% this file. If not, see <http://www.opensource.org/licenses/NPOSL-3.0>
%
% Calculation of the fluid flow in a tube.

classdef Fluid < handle
    properties (Constant)
        r_o=0.003;
        h=0.0003;
        t_b=1.0;
        sca=1.0e5;
        r_d=1.768e9;
        r_p=2.834e8;
        c_o=6.35e-10;
    end
    properties (SetAccess=immutable)
        dt;
        dz;
        m_e;
        m_fo;
        m_fg;
        m_fh;
        m_sg;
        c;
        pdcpds_1;
        s;
        rho;
        
        C_f;
        D_f;
        M_f;
        N_f;
    end
    properties (SetAccess=private)
        n;
        
        b_f;
        x_f;
        x_fn;
        x_sg;
        x_sgn;
        z_sg;
        z_fg;
        
        pdjpdx_f;
        a_f;
        a_fn;
        C_stxa_sg;
        D_stxa_sgn;
    end
    methods
        function F=Fluid(dt,dz,m_e,s,rho)
            F.dt=dt;
            F.dz=dz;
            F.m_e=m_e;
            F.m_fo=4+F.m_e;
            F.m_fg=F.m_e;
            F.m_fh=F.m_e;
            F.m_sg=F.m_e;
            F.c=F.c_o/(1.0+s(m_e+1,1)*0.5);
            F.pdcpds_1=-F.c_o/(1.0+s(m_e+1,1)*0.5)^2*0.5;
            F.s=s;
            F.rho=rho;
            
            F.C_f=(2.0*F.dz)/(F.dt*F.r_o)*eye(F.m_fh,F.m_sg);
            F.D_f=F.C_f;
            F.M_f=zeros(F.m_fo+F.m_fg);
            F.M_f(1,1)=1.0;
            F.M_f(2,2)=1.0;
            F.M_f(2,F.m_fo+1)=-2.0;
            F.M_f(2,F.m_fo+2)=1.0;
            F.M_f(3,3)=1.0;
            F.M_f(3,F.m_fo)=-2.0;
            F.M_f(3,F.m_fo-1)=1.0;
            F.M_f(4,3)=pi*F.r_o^2*(F.r_d+F.r_p+F.r_p*F.r_d*F.c/F.dt)/F.sca;
            F.M_f(4,4)=-(1.0+F.r_d*F.c/F.dt)/F.sca;
            F.M_f(5,2)=-0.5/F.rho;
            F.M_f(F.m_fo,4)=0.5/F.rho;
            F.M_f(5:F.m_fo,5:F.m_fo)=F.dz/F.dt*eye(F.m_fo-4);
            F.M_f(5:F.m_fo,F.m_fo+1:F.m_fo+F.m_fg)=0.5/F.rho...
                *(diag(ones(F.m_fg-1,1),1)-diag(ones(F.m_fg-1,1),-1));
            F.M_f(F.m_fo+1,1)=-0.5;
            F.M_f(F.m_fo+1,2)=-F.dt/F.dz/F.rho;
            F.M_f(F.m_fo+F.m_fg,3)=0.5;
            F.M_f(F.m_fo+F.m_fg,4)=-F.dt/F.dz/F.rho;
            F.M_f(F.m_fo+1:F.m_fo+F.m_fg,5:F.m_fo)=0.5...
                *(diag(ones(F.m_fo-5,1),1)-diag(ones(F.m_fo-5,1),-1));
            F.M_f(F.m_fo+1:F.m_fo+F.m_fg,F.m_fo+1:F.m_fo+F.m_fg)...
                =F.dt/F.dz/F.rho*(2.0*eye(F.m_fg)...
                -diag(ones(F.m_fg-1,1),1)-diag(ones(F.m_fg-1,1),-1));
            F.N_f=zeros(F.m_fo+F.m_fg);
            F.N_f(4,3)=pi*F.r_o^2*F.r_p*F.r_d*F.c/F.dt/F.sca;
            F.N_f(4,4)=-F.r_d*F.c/F.dt/F.sca;
            F.N_f(5:F.m_fo,5:F.m_fo)=F.dz/F.dt*eye(F.m_fo-4);
            
            F.n=0;
            
            F.x_f=zeros(F.m_fo+F.m_fg,1);
            F.x_fn=F.x_f;
            F.x_sg=zeros(F.m_sg,1);
            F.x_sgn=F.x_sg;
            F.z_sg=dz/2:dz:m_e*dz;
            F.z_fg=F.z_sg;
            
            F.a_f=zeros(F.m_fo+F.m_fg,1);
            F.a_fn=F.a_f;
            F.C_stxa_sg=zeros(F.m_fg,1);
            F.D_stxa_sgn=F.C_stxa_sg;
        end
        function x_fg=calculate_primal(F,x_sg)
            F.x_sg=x_sg;
            F.x_f=F.M_f\(F.b_f+F.N_f*F.x_fn...
                +[zeros(F.m_fo+F.m_fg-F.m_fh,1); F.D_f*F.x_sgn]...
                -[zeros(F.m_fo+F.m_fg-F.m_fh,1); F.C_f*F.x_sg]);
            x_fg=F.x_f(F.m_fo+1:F.m_fo+F.m_fg,1);
        end
        function C_ftxa_fg=calculate_dual(F,C_stxa_sg)
            F.C_stxa_sg=C_stxa_sg;
            F.a_f=F.M_f'\(-F.pdjpdx_f'+F.N_f'*F.a_fn...
                +[zeros(F.m_fo+F.m_fg-F.m_fg,1); F.D_stxa_sgn]...
                -[zeros(F.m_fo+F.m_fg-F.m_fg,1); F.C_stxa_sg]);
            C_ftxa_fg=F.C_f'*F.a_f(F.m_fo+F.m_fg-F.m_fh+1:F.m_fo+F.m_fg,1);
        end
        function D_ftxa_fgn=calculate_rhs_dual(F)
            D_ftxa_fgn=F.D_f'*F.a_fn(F.m_fo+F.m_fg-F.m_fh+1:F.m_fo+F.m_fg);
        end
        function adjust_rhs_dual(F,pdjpdx_f,D_stxa_sgn)
            F.pdjpdx_f=pdjpdx_f;
            F.D_stxa_sgn=D_stxa_sgn;
        end
        function pdR_fpds=calculate_gradient(F)
            pdR_fpds=zeros(F.m_fo+F.m_fg,length(F.s));
            pdR_fpds(4,length(F.s))=F.pdcpds_1*( ...
                pi*F.r_o^2*F.r_p*F.r_d/F.dt*(F.x_f(3)-F.x_fn(3)) ...
                -F.r_d/F.dt*(F.x_f(4)-F.x_fn(4)))/F.sca;
        end
        function v=inlet_velocity(F)
            t=F.n*F.dt/F.t_b;
            v=0.23+0.21*sin(2.0*pi*t) ...
                +0.11*cos(4.0*pi*(t-0.2)) ...
                +0.07*cos(6.0*pi*(t-0.2));
        end
        function increase_time(F)
            F.n=F.n+1;
            
            F.b_f=zeros(F.m_fo+F.m_fg,1);
            F.b_f(1,1)=F.inlet_velocity();
            F.x_fn=F.x_f;
            F.x_sgn=F.x_sg;
        end
        function decrease_time(F)
            F.n=F.n-1;
            
            F.pdjpdx_f=[];
            F.a_fn=F.a_f;
            F.D_stxa_sgn=[];
        end
    end
end