% Copyright (c) 2011 Joris Degroote, Majid Hojjat, Electra Stavropoulou,
% Roland Wüchner, Kai-Uwe Bletzinger
%
% Licensed under the Non-Profit Open Software License version 3.0
% (NPOSL-3.0). You should have received a copy of the NPOSL-3.0 along with 
% this file. If not, see <http://www.opensource.org/licenses/NPOSL-3.0>
%
% Quasi-Newton unconstrained optimiser with line search.

function [x,f,exitflag,output,grad,hess]=Optimiser(fun,x0,options)
    defaultopt=struct( ...
        'TolX',1e-6, ...
        'TolFun',1e-6, ...
        'MaxIter',400, ...
        'MaxFunEvals',1000, ...
        'MaxNodes',20, ...
        'OutputFcn',[] ...
    );
    if (nargin<3)
        options=defaultopt;
    end

    tolX=optimget(options,'TolX',defaultopt,'fast');
    tolFun=optimget(options,'TolFun',defaultopt,'fast');
    maxIter=optimget(options,'MaxIter',defaultopt,'fast');
    maxFunEvals=optimget(options,'MaxFunEvals',defaultopt,'fast');
    maxNodes=optimget(options,'MaxNodes',defaultopt,'fast');
    outputFcn=optimget(options,'OutputFcn',defaultopt,'fast');
        
    x=x0;
    funcCount=0;
    iterations=0;
    hess=HessianInverse(maxNodes);
    
    [f,grad]=fun(x0);
    funcCount=funcCount+1;
    g0Norm=norm(grad,Inf);
    hess.add(x0,grad');
    
    converged=false;
    exitflag=[];
    DisplayConvergence(iterations,funcCount,x,f,grad,0,outputFcn);
    
    while (~converged)    
        iterations=iterations+1;
        
        dir=-(hess*grad');
        dirDerivative=grad*dir; 
        
        [a,f,grad,funcCountLnSrch]=LineSearch(fun,x,dir,f,dirDerivative);
        funcCount=funcCount+funcCountLnSrch;
    
        deltaX=a*dir;
        x=x+deltaX;
        
        [converged,exitflag]=CheckConvergence(x,deltaX,iterations, ...
            funcCount,tolX,tolFun,maxIter,maxFunEvals,grad,g0Norm);
        DisplayConvergence(iterations,funcCount,x,f,grad,a,outputFcn);
        
        if (~converged)
            hess.add(x,grad');
        end
    end
    
    output=struct('iterations',iterations,'funcCount',funcCount);
end

function DisplayConvergence(iterations,funcCount,x,f,grad,a,outputFcn)
    headerstr=['  iter  fun. evals       fun. value    step length    ' ...
        ' optimality\n'];
    formatstr=' %5.0f       %5.0f    %13.6g  %13.6g   %12.3g\n';

    if (~mod(iterations,20))
        fprintf(headerstr);
    end
    fprintf(formatstr,iterations,funcCount,f,a,norm(grad,inf));
    
    if (~isempty(outputFcn))
        optimValues=struct('iteration',iterations,'fval',f, ...
            'funccount',funcCount,'stepsize',a, ...
            'firstorderopt',norm(grad,Inf));
        if (iterations)
            state='iter';
        else
            state='init';
        end
        for fcn=outputFcn
            stop=fcn(x,optimValues,state);
            if (stop)
                error('Output function requested');
            end
        end
    end
end

function [converged,exitflag]=CheckConvergence(x,deltaX,iter, ...
    funcCount,tolX,tolFun,maxIter,maxFunEvals,grad,g0Norm)

    if (norm(grad,Inf)<tolFun*(1+g0Norm))
        converged=true;
        exitflag=1;
    elseif (norm(deltaX./(1.0+abs(x)),inf)<tolX)
        converged=true;
        exitflag=2;
    elseif (funcCount>=maxFunEvals)
        converged=true;
        exitflag=0;
    elseif (iter>maxIter)
        converged=true;
        exitflag=0;
    else
        converged=false;
        exitflag=[];
    end
end

function [a,f,grad,funcCountLnSrch]=LineSearch(fun,x,dir,f,dirDerivative)
    c_1=1.0e-4;
    c_2=0.9;

    phi_0=f;
    dphida_0=dirDerivative;
    
    a_im=0.0;
    phi_im=phi_0;
    dphida_im=dphida_0;
    
    a_i=1.0;
    
    i=1;
    funcCountLnSrch=0;
    while (true)
        [f,grad]=fun(x+a_i*dir);
        funcCountLnSrch=funcCountLnSrch+1;
        phi_i=f;
        dphida_i=grad*dir;
        
        if ((phi_i>phi_0+c_1*a_i*dphida_0)||((phi_i>phi_im)&&(i>1)))
            [a,f,grad,funcCountZoom]=Zoom(fun,a_im,a_i,phi_im,phi_i, ...
                dphida_im,dphida_i,x,dir,phi_0,dphida_0,c_1,c_2);
            funcCountLnSrch=funcCountLnSrch+funcCountZoom;
            break;
        end
        
        if (abs(dphida_i)<=-c_2*dphida_0)
            a=a_i;
            break;
        end
        
        if (dphida_i>=0.0)
            [a,f,grad,funcCountZoom]=Zoom(fun,a_i,a_im,phi_i,phi_im, ...
                dphida_i,dphida_im,x,dir,phi_0,dphida_0,c_1,c_2);
            funcCountLnSrch=funcCountLnSrch+funcCountZoom;
            break;
        end
        
        a_im=a_i;
        phi_im=phi_i;
        dphida_im=dphida_i;
        
        a_i=a_i/2.0;
        i=i+1;
    end
end

function [a,f,grad,funcCountZoom]=Zoom(fun,a_lo,a_hi,phi_lo,phi_hi, ...
    dphida_lo,dphida_hi,x,dir,phi_0,dphida_0,c_1,c_2)
    
    funcCountZoom=0;
    while (true)
        d_1=dphida_lo+dphida_hi-3.0*(phi_lo-phi_hi)/(a_lo-a_hi);
        if (d_1^2-dphida_lo*dphida_hi<0)
            a_j=a_lo-(phi_hi-phi_lo)/dphida_lo;
        else
            d_2=sign(a_hi-a_lo)*sqrt(d_1^2-dphida_lo*dphida_hi);
            a_j=a_hi-(a_hi-a_lo)*(dphida_hi+d_2-d_1)/ ...
                (dphida_hi-dphida_lo+2.0*d_2);
        end
        [f,grad]=fun(x+a_j*dir);
        phi_j=f;
        dphida_j=grad*dir;
        funcCountZoom=funcCountZoom+1;
        if ((phi_j>phi_0+c_1*a_j*dphida_0)||(phi_j>=phi_lo))
            a_hi=a_j;
            phi_hi=phi_j;
            dphida_hi=dphida_j;
        else
            if (abs(dphida_j)<=-c_2*dphida_0)
                a=a_j;
                break;
            end
            if (dphida_j*(a_hi-a_lo)>=0.0)
                a_hi=a_lo;
                phi_hi=phi_lo;
                dphida_hi=dphida_lo;
            end
            a_lo=a_j;
            phi_lo=phi_j;
            dphida_lo=dphida_j;
        end
    end
end